/*
 *	bookmark.c
 */

#include "filer.h"

/*
 *	IsValidBMHead
 */
static BOOL IsValidBMHead(PBMHEAD pbh)
{
	if (memcmp((u8 *)pbh->sig, BMH_SIG, sizeof(pbh->sig)) ||
		pbh->ver != BMH_VER ||
		pbh->size != sizeof(*pbh))
		return FALSE;
	return TRUE;
}

/*
 *	ComputeFileSum
 */
static u16 ComputeFileSum(u8 *pname)
{
	u16		sum;
	u8		c;

	sum = 0;
	while (c = *pname++)
		sum += c;

	return sum;
}

/*
 *	InitBookMark
 */
void InitBookMark()
{
	PBMHEAD		pbh;
	PBMENTRY	pbm;
	int		i;

	pbh = (PBMHEAD)BOOKMARK_ADDR;
	if (!IsValidBMHead(pbh)) {
		/*
		 *	Build bookmark
		 */
		WriteSRAMByte(pbh->sig, BMH_SIG, sizeof(pbh->sig));
		WriteSRAM8(&pbh->ver, BMH_VER);
		WriteSRAM8(&pbh->size, sizeof(*pbh));
		WriteSRAM8(&pbh->nitems, MAX_BOOKMARKS);
		WriteSRAM8(&pbh->curindex, 0);

		pbm = (PBMENTRY)(BOOKMARK_ADDR + sizeof(*pbh));
		for (i = 0; i < MAX_BOOKMARKS; i++) {
			WriteSRAM8(&pbm->flags, 0);
			pbm++;
		}
	}
}

/*
 *	FindBookMark
 */
BOOL FindBookMark(PBMPARAMS parms)
{
	PBMHEAD		pbh;
	PBMENTRY	pbm, pbm1, pbm2;
	PDIRENT		pdir;
	int		i;
	u8		d;
	u16		sum;

	pbh = (PBMHEAD)BOOKMARK_ADDR;
	if (!IsValidBMHead(pbh))
		return FALSE;

	pdir = parms->pdir;
	i = (pbh->curindex > 0) ? pbh->curindex - 1 : 255;
	d = 255 - i;
	sum = ComputeFileSum(pdir->fileaddr);
	pbm = (PBMENTRY)(BOOKMARK_ADDR + sizeof(*pbh));
	pbm1 = NULL;
	pbm2 = pbm;

	for (i = 0; i < pbh->nitems; i++, pbm++) {
		if (pbm->flags & BMF_INUSE) {
			if (CompSRAMByte(pbm->fontname, parms->fontname,
				sizeof(pbm->fontname)) == 0 &&
				pbm->fontsize == parms->fontsize &&
				!CompSRAM16(pbm->filesum, &sum) &&
				!CompSRAM32(pbm->filesize, &pdir->filesize))
			{
				parms->pageno = ReadSRAM32(pbm->pageno);
				parms->offset = ReadSRAM32(pbm->offset);
				parms->pbm = pbm;
				return TRUE;
			}
			if ((u8)(pbm->index + d) < (u8)(pbm2->index + d))
				pbm2 = pbm;
		} else {
			if (!pbm1)
				pbm1 = pbm;
		}
	}

	parms->pbm = pbm1 ? pbm1 : pbm2;

	return FALSE;
}

/*
 *	BuildBookMark
 */
BOOL BuildBookMark(PBMPARAMS parms)
{
	PBMHEAD		pbh;
	PBMENTRY	pbm;
	PDIRENT		pdir;
	u16		sum;
	u8		i;
	BMPARAMS	parms1;

	pbh = (PBMHEAD)BOOKMARK_ADDR;
	if (!IsValidBMHead(pbh))
		return FALSE;

	pdir = parms->pdir;
	sum = ComputeFileSum(pdir->fileaddr);

	parms1 = *parms;
	(void)FindBookMark(&parms1);
	pbm = parms1.pbm;

	WriteSRAM8(&pbm->flags, BMF_INUSE);
	WriteSRAM8(&pbm->index, pbh->curindex);
	WriteSRAM16(pbm->filesum, sum);
	WriteSRAM32(pbm->filesize, pdir->filesize);
	WriteSRAM32(pbm->pageno, parms->pageno);
	WriteSRAM32(pbm->offset, parms->offset);
	WriteSRAMByte(pbm->fontname, parms->fontname, sizeof(pbm->fontname));
	WriteSRAM8(&pbm->fontsize, parms->fontsize);

	if (pbh->curindex >= 255)
		pbh->curindex = 0;
	else
		pbh->curindex++;

	return TRUE;
}

