/*
 *	filer.h
 */

#ifndef _FILER_H_
#define _FILER_H_

#include <gba.h>
#include <font8.h>
#include <gbafs.h>
#include <xfont.h>
#include <chartype.h>

#include "bookmark.h"
#include "bitmap.h"
#include "fonts.h"
#include "listbox.h"
#include "timer.h"

/*
 *	Constants
 */

#define COLOR_WHITE	RGB(31,31,31)
#define COLOR_BLACK	RGB(0,0,0)

#define KEY_REPEAT	0x4000

/*
 *	SRAM space definitions
 */

#define BOOKMARK_ADDR	SRAM_ADDR
#define BOOKMARK_SIZE	(sizeof(BMHEAD) + sizeof(BMENTRY) * MAX_BOOKMARKS)

#define FONTSIZE_ADDR	(BOOKMARK_ADDR + BOOKMARK_SIZE)

/*
 *	Structures
 */

typedef struct _BITMAP {
	int		width;
	int		height;
	const u16	*pimage;
} BITMAP, *PBITMAP;

/*
 *	Globals
 */

extern const BITMAP	bmFolder;
extern const BITMAP	bmBookmark;

/*
 *	External functions
 */

/* sram.c */
void ReadSRAMByte(vu8 *sram, u8 *pv, int count);
void WriteSRAMByte(vu8 *sram, u8 *pv, int count);
u8 ReadSRAM8(vu8 *sram);
u32 ReadSRAM32(vu8 *sram);
void WriteSRAM8(vu8 *sram, u8 v);
void WriteSRAM16(vu8 *sram, u16 v);
void WriteSRAM32(vu8 *sram, u32 v);
int CompSRAM16(vu8 *sram, u16 *pv);
int CompSRAM32(vu8 *sram, u32 *pv);
int CompSRAMByte(vu8 *sram, u8 *pv, int count);

/* subr.c */
void ClearScreen();
void FillScreen(int x, int y, int width, int height, u16 c);
void EraseScreen(int x, int y, int width, int height);
void ReverseScreen(int x, int y, int width, int height);
void CopyBitmap16(PBITMAP pbm, int x, int y);
int GetKey();
int PutStr(int x, int y, u8 *s);
int Printf(int x, int y, u8 *fmt, ...);
int LeftPrintf(int x, int y, int w, u8 *fmt, ...);
int RightPrintf(int x, int y, int w, u8 *fmt, ...);

/* timer.c*/
void InitTimer();
void Wait(u32 msec);

/* view.c */
void StartView(PDIRENT pdir);

/* viewtext.c */
void ViewText(PDIRENT pdir);

#endif // _FILER_H_

