/*
 *	listbox.c
 */

#include "filer.h"

static int	Xlist, Ylist, Wlist, Hlist;

/*
 *	InitListbox
 */
void InitListBox(int x, int y, int w, int h)
{
	Xlist = x;
	Ylist = y;
	Wlist = w;
	Hlist = h;
}

/*
 *	CharWidths
 */
static int CharWidths(u8 *s)
{
	int		w;

	w = 0;
	while (*s)
		w += XfontGetCharWidth(&s);

	return w;
}

static const char	HL[2] = { 0x84,0x9f };
static const char	VL[2] = { 0x84,0xa0 };
static const char	UL[2] = { 0x84,0xa1 };
static const char	UR[2] = { 0x84,0xa2 };
static const char	LR[2] = { 0x84,0xa3 };
static const char	LL[2] = { 0x84,0xa4 };

/*
 *	Listbox
 */
int ListBox(PLISTBOX plb)
{
	int		i, n, w, h;
	int		ix;
	int		wl, hl, ws, hs;
	int		x, y, x0, y0;
	u8		*p;
	int		key;

	XfontSelectFontObject(FO_SMALL);
	w = XfontGetMaxCharWidth(0x81);
	h = XfontGetCharHeight();

	/*
	 *	Compute the size and positions
	 */

	wl = CharWidths(plb->title);
	hl = h;
	for (i = 0; i < plb->nitems; i++) {
		n = CharWidths(plb->items[i]);
		if (n > wl)
			wl = n;
		hl += h;
	}
	wl = (wl + w - 1) / w * w;	// round up to kanji width

	ws = XfontGetMaxCharWidth(0x81) * 2;
	hs = h * 2;

	if (wl + ws > Wlist || hl + hs > Hlist)
		return LB_ERROR;

	x0 = Xlist + (Wlist - (wl + ws)) / 2;
	y0 = Ylist + (Hlist - (hl + hs)) / 2;

	/*
	 *	Put the listbox
	 */

	x = x0;
	y = y0;

	// Title

	p = (u8 *)UL;
	x += XfontPutChar(x, y, &p);

	LeftPrintf(x, y, wl, "%s", plb->title);
	x += wl;

	p = (u8 *)UR;
	x += XfontPutChar(x, y, &p);

	y += h;

	// Items

	for (i = 0; i < plb->nitems; i++) {
		x = x0;

		p = (u8 *)VL;
		x += XfontPutChar(x, y, &p);

		RightPrintf(x, y, wl, "%s", plb->items[i]);
		x += wl;

		p = (u8 *)VL;
		x += XfontPutChar(x, y, &p);

		y += h;
	}

	// Tail

	x = x0;

	p = (u8 *)LL;
	x += XfontPutChar(x, y, &p);

	n = wl / w;
	for (i = 0; i < n; i++) {
		p = (u8 *)HL;
		x += XfontPutChar(x, y, &p);
	}

	p = (u8 *)LR;
	x += XfontPutChar(x, y, &p);

	/*
	 *	Select item
	 */

	x0 += w;
	y0 += h;
	ix = plb->defitem;

	for (;;) {
		ReverseScreen(x0, y0 + (ix * h), wl, h);
		while (!(key = GetKey()))
			;
		ReverseScreen(x0, y0 + (ix * h), wl, h);

		if (key & (KEY_SELECT | KEY_B))
			return LB_CANCEL;
		else if (key & KEY_A)
			break;
		else if (key & KEY_UP) {
			if (ix > 0)
				ix--;
			else
				ix = plb->nitems - 1;
		} else if (key & KEY_DOWN) {
			if (ix < plb->nitems - 1)
				ix++;
			else
				ix = 0;
		}
	}

	return ix;
}

