/*
 *	view.c
 */

#include "filer.h"

#define X_FILER		2
#define Y_FILER		2
#define W_FILER		(SCREEN_WIDTH - X_FILER * 2)
#define H_FILER		(SCREEN_HEIGHT - Y_FILER * 2)

static int	Wsmall, Hsmall;

static int	Xpage, Ypage, Wpage, Hpage;
static int	Xpath, Ypath, Wpath, Hpath;
static int	Xsep, Ysep, Wsep, Hsep;
static int	Xfile, Yfile, Wfile, Hfile;
static int	Xicon, Yicon, Wicon, Hicon;
static int	Xname, Yname, Wname, Hname;
static int	nFilesPP;

/*
 *	view object
 */

typedef struct _VIEWOBJ {
	struct _VIEWOBJ	*ppv;
	PDIRENT		pdir;
	u8		*ppath;
	u8		*pcurpath;
	int		nfiles;
	int		npages;
	int		curpage;
	int		curindex;
} VIEWOBJ, *PVIEWOBJ;

//========================

/*
 *	InitPositions
 */
static void InitPositions()
{
	int		wk;

	XfontSelectFontObject(FO_SMALL);
	Wsmall = XfontGetMaxCharWidth(' ');
	Hsmall = XfontGetCharHeight();

	wk = XfontGetMaxCharWidth(0x81);

	Wpage = Wsmall * 13;	// (xxxxx/xxxxx)
	Hpage = Hsmall;
	Xpage = (X_FILER + W_FILER - Wpage);
	Ypage = Y_FILER;

	Wpath = (W_FILER - Wpage);
	Hpath = Hsmall;
	Xpath = X_FILER;
	Ypath = Y_FILER;

	Wsep = W_FILER;
	Hsep = 1;
	Xsep = X_FILER;
	Ysep = Ypath + Hsmall;

	Xfile = X_FILER;
	Yfile = Ysep + Hsep + 1;
	Wfile = W_FILER;
	Hfile = (Y_FILER + H_FILER - Yfile);

	Wicon = bmFolder.width;
	Hicon = bmFolder.height;
	Xicon = Xfile + (wk - bmFolder.width) / 2;
	Yicon = Yfile + (Hsmall - bmFolder.height) / 2;

	Xname = X_FILER + wk;
	Yname = Yfile;
	Wname = W_FILER - wk;
	Hname = Hsmall;

	nFilesPP = Hfile / Hsmall;
}

/*
 *	ViewInit
 */
static void ViewInit(PVIEWOBJ pv)
{
	PDIRENT		pdir;
	int		len;

	pdir = pv->pdir;
	if (pdir->namesize) {
		len = strlen(pdir->fileaddr);
		*pv->pcurpath++ = '\\';
		strcpy(pv->pcurpath, pdir->fileaddr);
		pv->pcurpath += len;
	} else
		pv->pcurpath[0] = '\0';
	pv->nfiles = pdir->filesize / sizeof(*pdir);
	pv->npages = (pv->nfiles + nFilesPP - 1) / nFilesPP;
	pv->curpage = 0;
	pv->curindex = 0;
}

/*
 *	ViewPathName
 */
static void ViewPathName(PVIEWOBJ pv)
{
	LeftPrintf(Xpath, Ypath, Wpath, "%s", pv->ppath);
}

/*
 *	ViewPageNumber
 */
static void ViewPageNumber(PVIEWOBJ pv)
{
	if (pv->npages > 1)
		RightPrintf(Xpage, Ypage, Wpage, "(%d/%d)",
			pv->curpage + 1, pv->npages);
	else
		EraseScreen(Xpage, Ypage, Wpage, Hpage);
}

/*
 *	ViewFiles
 */
static void ViewFiles(PVIEWOBJ pv)
{
	int		i, n, x, y, yi;
	PDIRENT		pdir;

	y = Yname;
	yi = Yicon;
	i = pv->curpage * nFilesPP;
	n = min(pv->nfiles, nFilesPP + i);
	pdir = (PDIRENT)(pv->pdir->fileaddr + pv->pdir->namesize) + i;

	ViewPageNumber(pv);

	for (; i < n; i++, pdir++) {
		if (pdir->attr & DA_DIR)
			CopyBitmap16((PBITMAP)&bmFolder, Xicon, yi);
		else
			EraseScreen(Xicon, yi, Wicon, Hicon);
		LeftPrintf(Xname, y, Wname, "%s", pdir->fileaddr);
		y += Hname;
		yi += Hname;
	}

	EraseScreen(Xfile, y, Wfile, (Yfile + Hfile) - y);
}

/*
 *	ViewNewPage
 */
static void ViewNewPage(PVIEWOBJ pv)
{
	int		p;

	p = pv->curindex / nFilesPP;
	if (p != pv->curpage) {
		pv->curpage = p;
		ViewFiles(pv);
	}
}

/*
 *	ViewUp
 */
static void ViewUp(PVIEWOBJ pv)
{
	if (!pv->nfiles)
		return;
	if (pv->curindex <= 0)
		pv->curindex = pv->nfiles - 1;
	else
		pv->curindex--;
	ViewNewPage(pv);
}

/*
 *	ViewDown
 */
static void ViewDown(PVIEWOBJ pv)
{
	if (!pv->nfiles)
		return;
	if (pv->curindex >= pv->nfiles - 1)
		pv->curindex = 0;
	else
		pv->curindex++;
	ViewNewPage(pv);
}

/*
 *	ViewLeft
 */
static void ViewLeft(PVIEWOBJ pv)
{
	if (!pv->nfiles)
		return;
	if (pv->npages <= 1) {
		pv->curindex = 0;
		return;
	}
	if (pv->curpage <= 0)
		pv->curpage = pv->npages - 1;
	else
		pv->curpage--;
	pv->curindex = pv->curpage * nFilesPP;
	ViewFiles(pv);
}

/*
 *	ViewRight
 */
static void ViewRight(PVIEWOBJ pv)
{
	if (!pv->nfiles)
		return;
	if (pv->npages <= 1) {
		pv->curindex = pv->nfiles - 1;
		return;
	}
	if (pv->curpage >= pv->npages - 1)
		pv->curpage = 0;
	else
		pv->curpage++;
	pv->curindex = pv->curpage * nFilesPP;
	ViewFiles(pv);
}

/*
 *	ViewSeparator
 */
static void ViewSeparator()
{
	FillScreen(Xsep, Ysep, Wsep, Hsep, RGB(15,15,15));
}

/*
 *	RefreshView
 */
static void RefreshView(PVIEWOBJ pv, BOOL force)
{
	XfontSelectFontObject(FO_SMALL);
	if (force) {
		ClearScreen();
		ViewSeparator();
	}
	ViewPathName(pv);
	ViewFiles(pv);
}

static void DoView(PVIEWOBJ pv);

/*
 *	ViewEnter
 */
static void ViewEnter(PVIEWOBJ pv)
{
	PDIRENT		pdir;

	if (!pv->nfiles)
		return;
	pdir = (PDIRENT)(pv->pdir->fileaddr + pv->pdir->namesize) +
		pv->curindex;
	if (pdir->attr & DA_DIR) {
		VIEWOBJ		vo;

		vo.ppv = pv;
		vo.ppath = pv->ppath;
		vo.pcurpath = pv->pcurpath;
		vo.pdir = pdir;
		DoView(&vo);
		*pv->pcurpath = '\0';
		RefreshView(pv, FALSE);
	} else {
		if (!ViewBMP(pdir))
			ViewText(pdir);
		RefreshView(pv, TRUE);
	}
}

/*
 *	DoView
 */
static void DoView(PVIEWOBJ pv)
{
	int		y;
	int		key;

	ViewInit(pv);
	RefreshView(pv, FALSE);

	for (;;) {
		y = (pv->curindex - pv->curpage * nFilesPP) * Hname + Yname;
		ReverseScreen(Xname, y, Wname, Hname);
		while ((key = GetKey()) == 0)
			;
		ReverseScreen(Xname, y, Wname, Hname);
		if (key & KEY_UP)
			ViewUp(pv);
		else if (key & KEY_DOWN)
			ViewDown(pv);
		else if (key & KEY_LEFT)
			ViewLeft(pv);
		else if (key & KEY_RIGHT)
			ViewRight(pv);
		else if (key & KEY_A)
			ViewEnter(pv);
		else if (key & KEY_B) {
			if (pv->ppv)
				break;
		}
	}
}

/*
 *	StartView
 */
void StartView(PDIRENT pdir)
{
	VIEWOBJ		vo;
	u8		pathbuf[256];

	ClearScreen();
	InitPositions();
	ViewSeparator();

	vo.ppath = pathbuf;
	vo.pcurpath = vo.ppath;
	vo.ppv = NULL;
	vo.pdir = pdir;

	DoView(&vo);
}

