/*
 *	viewbmp.c
 */

#include "filer.h"

#define X_ERROR		80
#define Y_ERROR		20

#define MAKEU32(w)	(w[0] | (w[1] << 16))
#define RGB16(r, g, b)	\
	((((b) << 7) & 0x7c00) | (((g) << 2) & 0x3e0) | (((r) >> 3) & 0x1f))

/*
 *	IsBitmap
 */
static BOOL IsBitmap(u8 *pfile)
{
	return memcmp(pfile, BMFH_SIG, 2) == 0;
}

/*
 *	ShowBMP
 */
static void ShowBMP(int x, int y, u8 *pbuf)
{
	PBMPFILEHDR	pbmf;
	PBMPINFOHDR	pbmi;
	u8		*bp0, *bp;
	vu16		*vp;
	u8		r, g, b;
	int		width, height;
	int		wsize;

	if (!IsBitmap(pbuf))
		return;

	pbmf = (PBMPFILEHDR)pbuf;
	pbmi = (PBMPINFOHDR)(pbuf + SIZEOF_BMPFILEHDR);
	width = MAKEU32(pbmi->biWidth) - x;
	if (width <= 0)
		return;
	height = MAKEU32(pbmi->biHeight) - y;
	if (height <= 0)
		return;

	width = min(width, SCREEN_WIDTH);
	height = min(height, SCREEN_HEIGHT);
	wsize = (MAKEU32(pbmi->biWidth) * 3 + 3) & ~3;

	bp0 = &pbuf[MAKEU32(pbmf->bfSize) - y * wsize + x * 3];
	vp = (vu16 *)VRAM_ADDR;

	for (y = 0; y < height; y++) {
		bp0 -= wsize;
		bp = bp0;
		for (x = 0; x < width; x++) {
			b = *bp++;
			g = *bp++;
			r = *bp++;
			*vp++ = RGB16(r, g, b);
		}
		while (x++ < SCREEN_WIDTH)
			*vp++ = 0;
	}
	if (y < SCREEN_HEIGHT)
		EraseScreen(0, y, SCREEN_WIDTH, SCREEN_HEIGHT - y);
}

/*
 *	ViewBMP
 */
BOOL ViewBMP(PDIRENT pdir)
{
	PBMPFILEHDR	pbmf;
	PBMPINFOHDR	pbmi;
	int		key;
	u32		wsize, hsize;
	int		width, height;
	int		curx, cury;
	int		dif;

	if (!pdir->filesize)
		return FALSE;

	pbmf = (PBMPFILEHDR)(pdir->fileaddr + pdir->namesize);
	if (!IsBitmap((u8 *)pbmf) || MAKEU32(pbmf->bfSize) != pdir->filesize)
		return FALSE;

	pbmi = (PBMPINFOHDR)((u8 *)pbmf + SIZEOF_BMPFILEHDR);
	if (MAKEU32(pbmi->biSize) != sizeof(*pbmi))
		return FALSE;

	if (pbmi->biBitCount != 24)
		// TODO: Support another color modes.
		return FALSE;

	width = MAKEU32(pbmi->biWidth);
	wsize = (width * 3 + 3) & ~3;	// round up to 4 bytes
	hsize = MAKEU32(pbmi->biHeight);
	height = hsize;

	curx = cury = 0;

	ShowBMP(curx, cury, (u8 *)pbmf);

	for (;;) {
		key = GetKey();
		dif = (key & KEY_REPEAT) ? 32 : 8;

		if (key & KEY_B)
			break;
		else if (key & KEY_LEFT) {
			if ((curx - dif) >= 0) {
				curx -= dif;
				ShowBMP(curx, cury, (u8 *)pbmf);
			}
		} else if (key & KEY_RIGHT) {
			if ((curx + SCREEN_WIDTH + dif) < width) {
				curx += dif;
				ShowBMP(curx, cury, (u8 *)pbmf);
			}
		} else if (key & KEY_UP) {
			if ((cury - dif) >= 0) {
				cury -= dif;
				ShowBMP(curx, cury, (u8 *)pbmf);
			}
		} else if (key & KEY_DOWN) {
			if ((cury + SCREEN_HEIGHT + dif) < height) {
				cury += dif;
				ShowBMP(curx, cury, (u8 *)pbmf);
			}
		}
	}

	return TRUE;
}
