/*
 *	viewtext.c
 */

#include "filer.h"

#define X_TEXTVIEW	2
#define Y_TEXTVIEW	2
#define W_TEXTVIEW	(SCREEN_WIDTH - X_TEXTVIEW * 2)
#define H_TEXTVIEW	(SCREEN_HEIGHT - Y_TEXTVIEW * 2)

#define INVALID_PAGENO	(0xffffffff)
#define INVALID_OFFSET	(0xffffffff)

/*
 *	variables
 */

static int	Wsmall, Hsmall;
static int	Wnormal, Hnormal;

static int	Xpage, Ypage, Wpage, Hpage;
static int	Xmark, Ymark, Wmark, Hmark;
static int	Xname, Yname, Wname, Hname;
static int	Xsep, Ysep, Wsep, Hsep;
static int	Xtext, Ytext, Wtext, Htext;

static u32	CurPage;
static u32	BookmarkPage;

static u32	*pPage;
static u32	*pPageEnd;

static u32	nPages;
static u32	workPage;
static u32	workOffset;
static u32	StartPage;
static u32	StartOffset;
static u8	PageStat;

#define PS_FORW		0x01
#define PS_BACK		0x02
#define PS_DONE		0x03

static u8	*pTextName;
static u8	*pText;
static u32	nText;

/*
 *	InitPositions
 */
static void InitPositions()
{
	int		wk;

	XfontSelectFontObject(FO_SMALL);
	Wsmall = XfontGetMaxCharWidth(' ');
	Hsmall = XfontGetCharHeight();

	wk = XfontGetMaxCharWidth(0x81);

	XfontSelectFontObject(FO_NORMAL);
	Wnormal = XfontGetMaxCharWidth(' ');
	Hnormal = XfontGetCharHeight();

	Wpage = Wsmall * 13;	// (xxxxx/xxxxx)
	Hpage = Hsmall;
	Xpage = (X_TEXTVIEW + W_TEXTVIEW - Wpage);
	Ypage = Y_TEXTVIEW;

	Wmark = bmBookmark.width;
	Hmark = bmBookmark.height;
	Xmark = X_TEXTVIEW + (wk - bmBookmark.width) / 2;
	Ymark = Y_TEXTVIEW + (Hsmall - bmBookmark.height) / 2;

	Wname = (W_TEXTVIEW - wk - Wpage);
	Hname = Hsmall;
	Xname = X_TEXTVIEW + wk;
	Yname = Y_TEXTVIEW;

	Wsep = W_TEXTVIEW;
	Hsep = 1;
	Xsep = X_TEXTVIEW;
	Ysep = Yname + Hname;

	Xtext = X_TEXTVIEW;
	Ytext = Ysep + Hsep + 1;
	Wtext = W_TEXTVIEW;
	Htext = (Y_TEXTVIEW + H_TEXTVIEW - Ytext);
}

/*
 *	ViewName
 */
static void ViewName()
{
	XfontSelectFontObject(FO_SMALL);
	LeftPrintf(Xname, Yname, Wname, "%s", pTextName);
}

/*
 *	ViewBookmark
 */
static void ViewBookmark()
{
	if (BookmarkPage != INVALID_PAGENO && BookmarkPage == CurPage)
		CopyBitmap16((PBITMAP)&bmBookmark, Xmark, Ymark);
	else
		EraseScreen(Xmark, Ymark, Wmark, Hmark);
}

/*
 *	ViewPageNumber
 */
static void ViewPageNumber()
{
	u32		pageno;

	if (nPages > 1) {
		if (PageStat == PS_DONE)
			pageno = nPages;
		else
			pageno = workPage;
		XfontSelectFontObject(FO_SMALL);
		RightPrintf(Xpage, Ypage, Wpage, "(%d/%d)",
			CurPage + 1, pageno);
	} else
		EraseScreen(Xpage, Ypage, Wpage, Hpage);
}

/*
 *	InitPage
 */
static void InitPage(u32 offset)
{
	pPage = (u32 *)EWRAM_ADDR;
	pPageEnd = (u32 *)(EWRAM_ADDR + EWRAM_SIZE);

	StartPage = workPage = CurPage;
	StartOffset = workOffset = offset;
	nPages = 0;
	PageStat = PS_FORW;
}

/*
 *	IsStopPage
 */
static BOOL IsStopPage()
{
	switch (PageStat) {
	case PS_FORW:
		if (workOffset >= nText)
			return TRUE;
		break;
	case PS_BACK:
		if (workOffset >= StartOffset)
			return TRUE;
		break;
	}

	return FALSE;
}

/*
 *	SetupPage
 */
static BOOL SetupPage()
{
	int		x, y, w;
	int		wk;
	u8		c, c2;
	u8		*pt;
	static u32	ticks;

	if (PageStat == PS_DONE)
		return TRUE;

	if (GetKey() & KEY_B)
		return FALSE;

	if (&pPage[workPage] >= pPageEnd) {
		// TODO: Error handling
		XfontSelectFontObject(FO_SMALL);
		LeftPrintf(X_TEXTVIEW, Y_TEXTVIEW, W_TEXTVIEW, "No Memory.");
		return FALSE;
	}

	XfontSelectFontObject(FO_NORMAL);
	wk = XfontGetMaxCharWidth(0x81);

	pPage[workPage] = workOffset;

	for (y = Ytext; y < (Ytext + Htext - Hnormal); y += Hnormal) {
		x = Xtext;
		while (x < (Xtext + Wtext - wk)) {
			if (IsStopPage())
				goto nextpage;
			pt = &pText[workOffset];
			c = *pt;
			workOffset += GETCHARLENGTH(c);
			switch (c) {
			default:
				x += XfontGetCharWidth(&pt);
				break;
			case 0x0d:
				x = Xtext;
				pt++;
				break;
			case 0x0a:
				pt++;
				goto nextline;
			}
		}
	nextline:
		;
	}

nextpage:
	workPage++;
	switch (PageStat) {
	case PS_FORW:
		nPages = workPage;
		if (IsStopPage()) {
			if (StartPage) {
				workPage = 0;
				workOffset = 0;
				PageStat = PS_BACK;
			} else
				PageStat = PS_DONE;
		}
		break;
	case PS_BACK:
		if (IsStopPage())
			PageStat = PS_DONE;
		break;
	}

	if (PageStat == PS_DONE || (ClockTick > ticks && nPages > 1)) {
		ticks = ClockTick + 500;
		ViewPageNumber();
	}

	return TRUE;
}

/*
 *	IsPageReady
 */
static BOOL IsPageReady(u32 pageno)
{
	switch (PageStat) {
	case PS_DONE:
		return TRUE;
	case PS_BACK:
		if (pageno < workPage)
			return TRUE;
		if (pageno >= StartPage && pageno < nPages)
			return TRUE;
		break;
	case PS_FORW:
		if (pageno >= StartPage && pageno < workPage)
			return TRUE;
		break;
	}

	return FALSE;
}

/*
 *	GetPageOffset
 */
static u32 GetPageOffset(u32 pageno)
{
	if (!nText)
		return INVALID_PAGENO;

	while (!IsPageReady(pageno))
		if (!SetupPage())
			return INVALID_PAGENO;

	return pPage[pageno];
}

/*
 *	PutText
 */
static void PutText(int pageno)
{
	int		x, y, w, len;
	int		wk;
	u8		c;
	u32		offset;
	u8		*pt, *pe;

	offset = GetPageOffset(pageno);
	if (offset == INVALID_PAGENO)
		return;
	ViewBookmark();
	ViewPageNumber();

	XfontSelectFontObject(FO_NORMAL);
	wk = XfontGetMaxCharWidth(0x81);

	pt = &pText[offset];
	pe = &pText[nText];
	for (y = Ytext; y < (Ytext + Htext - Hnormal); y += Hnormal) {
		x = Xtext;
		len = 0;
		while (x < (Xtext + Wtext - wk)) {
			if (pt >= pe)
				goto out;
			c = *pt;
			switch (c) {
			default:
				w = XfontPutChar(x, y, &pt);
				x += w;
				len += w;
				break;
			case 0x0d:
				x = Xtext;
				pt++;
				break;
			case 0x0a:
				pt++;
				goto nextline;
			}
		}
	nextline:
		if (len < Wtext)
			EraseScreen(Xtext + len, y, Wtext - len, Hnormal);
	}

out:
	if (y < (Ytext + Htext))
		EraseScreen(Xtext, y, Wtext, (Ytext + Htext) - y);
}

/*
 *	ViewSeparator
 */
static void ViewSeparator()
{
	FillScreen(Xsep, Ysep, Wsep, Hsep, RGB(15,15,15));
}

/*
 *	RefreshView
 */
static void RefreshView(BOOL force)
{
	InitPositions();
	InitListBox(Xtext, Ytext, Wtext, Htext);
	if (force) {
		ClearScreen();
		ViewName();
		ViewSeparator();
	}
}

/*
 *	PageUp
 */
static BOOL PageUp(BOOL *pdirty)
{
	if (CurPage > 0) {
		CurPage--;
		*pdirty = TRUE;
		return TRUE;
	} else {
		while (PageStat != PS_DONE) {
			if (!SetupPage())
				return FALSE;
		}
		if (nPages <= 1) {
			*pdirty = FALSE;
			return TRUE;
		}
		CurPage = nPages - 1;
		*pdirty = TRUE;
		return TRUE;
	}
}

/*
 *	PageDown
 */
static BOOL PageDown(BOOL *pdirty)
{
	if (CurPage >= nPages) {
		while (!IsPageReady(CurPage)) {
			if (!SetupPage())
				return FALSE;
		}
	}
	if (PageStat == PS_DONE) {
		if (nPages <= 1) {
			*pdirty = FALSE;
			return TRUE;
		}
		if (CurPage + 1 >= nPages)
			CurPage = 0;
		else
			CurPage++;
	} else
		CurPage++;
	*pdirty = TRUE;
	return TRUE;
}

/*
 *	ViewText
 */
void ViewText(PDIRENT pdir)
{
	int		key;
	BOOL		dirty;
	BOOL		inrepeat;
	u32		offset;
	BMPARAMS	parms;

	pText = pdir->fileaddr + pdir->namesize;
	nText = pdir->filesize;
	pTextName = pdir->fileaddr;

setfont:
	XfontSelectFontObject(FO_NORMAL);
	parms.pdir = pdir;
	parms.fontname = XfontGetFontName();
	parms.fontsize = XfontGetCharHeight();
	if (FindBookMark(&parms)) {
		CurPage = parms.pageno;
		offset = parms.offset;
		BookmarkPage = CurPage;
	} else {
		CurPage = 0;
		offset = 0;
		BookmarkPage = INVALID_PAGENO;
	}

	InitPage(offset);

restart:
	RefreshView(TRUE);

	dirty = FALSE;
	inrepeat = FALSE;

	PutText(CurPage);

	for (;;) {
		while ((key = GetKey()) == 0) {
			if (!SetupPage())
				return;
			if (inrepeat) {
				PutText(CurPage);
				inrepeat = FALSE;
			}
		}

		if (key & KEY_B)
			break;
		else if (key & KEY_A) {
			CurPage = 0;
			goto restart;
		} else if (key & (KEY_RIGHT|KEY_DOWN)) {
			if (!PageDown(&dirty))
				return;
		} else if (key & (KEY_LEFT|KEY_UP)) {
			if (!PageUp(&dirty))
				return;
		} else if (key & KEY_L) {
			XfontSelectFontObject(FO_NORMAL);
			parms.pdir = pdir;
			parms.pageno = CurPage;
			parms.offset = GetPageOffset(CurPage);
			parms.fontname = XfontGetFontName();
			parms.fontsize = XfontGetCharHeight();
			if (BuildBookMark(&parms)) {
				BookmarkPage = CurPage;
				ViewBookmark();
			}
		} else if (key & KEY_SELECT) {
			if (SelectFont())
				goto setfont;
			dirty = TRUE;
		}

		if (dirty) {
			if (key & KEY_REPEAT) {
				ViewBookmark();
				ViewPageNumber();
				Wait(max(10, 2000 / nPages));
				inrepeat = TRUE;
			} else
				PutText(CurPage);
			dirty = FALSE;
		}
	}
}

