/*
 *	elisa.c
 */

#include <elisa.h>

#define FONT_WIDTH	9
#define FONT_HEIGHT	9

#define INDEX2OFFSET(ix)	((ix) << 3)
#define INVALID_CODE		0x8145
#define INVALID_INDEX		5
#define INVALID_OFFSET		INDEX2OFFSET(INVALID_INDEX)

/*
 *	structures
 */

typedef struct _FONTMAP {
	u8		start;
	u8		end;
	u16		index;
} FONTMAP, *PFONTMAP;

typedef struct _MFONTMAP {
	int		count;
	PFONTMAP	pfm;
} MFONTMAP, *PMFONTMAP;

typedef struct _FONTHAND {
	int		(*handler)(u8 cl, void *pdata);
	void		*pdata;
} FONTHAND, *PFONTHAND;

/*
 *	font handler
 */

int fhInvalid(u8 cl, void *pdata)
{
	return INVALID_INDEX;
}

int fhSingle(u8 cl, void *pdata)
{
	PFONTMAP	pfm = (PFONTMAP)pdata;

	if (cl < pfm->start || cl > pfm->end)
		return INVALID_INDEX;
	else
		return pfm->index + (cl - pfm->start);
}

int fhMulti(u8 cl, void *pdata)
{
	PMFONTMAP	pmf = (PMFONTMAP)pdata;
	PFONTMAP	pfm;
	int		i;

	for (i = 0, pfm = pmf->pfm; i < pmf->count; i++, pfm++) {
		if (cl > pfm->end || cl < pfm->start)
			continue;
		return pfm->index + (cl - pfm->start);
	}
	return INVALID_INDEX;
}

/*
 *	font map data
 */

#define ARRAYSIZE(a)	(sizeof(a) / sizeof(a[0]))

//===	81

static const FONTMAP	fm8140[] = {
	{ 0x40, 0x7e, 0x0000 },
};

static const FONTMAP	fm8180[] = {
	{ 0x80, 0xac, 0x003f },
	{ 0xb8, 0xbf, 0x006c },
};
static const MFONTMAP	mfm8180[] = { ARRAYSIZE(fm8180), (PFONTMAP)fm8180 };

static const FONTMAP	fm81c0[] = {
	{ 0xc8, 0xce, 0x0074 },
	{ 0xda, 0xe8, 0x007b },
	{ 0xf0, 0xf7, 0x008a },
	{ 0xfc, 0xfc, 0x0092 },
};
static const MFONTMAP	mfm81c0[] = { ARRAYSIZE(fm81c0), (PFONTMAP)fm81c0 };

static const FONTHAND	fh81[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fm8140 },
	{ fhMulti, (void *)mfm8180 },
	{ fhMulti, (void *)mfm81c0 },
};

//===	82

static const FONTMAP	fm8240[] = {
	{ 0x4f, 0x58, 0x0093 },
	{ 0x60, 0x79, 0x009d },
};
static const MFONTMAP	mfm8240[] = { ARRAYSIZE(fm8240), (PFONTMAP)fm8240 };

static const FONTMAP	fm8280[] = {
	{ 0x81, 0x9a, 0x00b7 },
	{ 0x9f, 0xbf, 0x00d1 },
};
static const MFONTMAP	mfm8280[] = { ARRAYSIZE(fm8280), (PFONTMAP)fm8280 };

static const FONTMAP	fm82c0[] = {
	{ 0xc0, 0xf1, 0x00f2 },
};

static const FONTHAND	fh82[] = {
	{ fhInvalid, NULL },
	{ fhMulti, (void *)mfm8240 },
	{ fhMulti, (void *)mfm8280 },
	{ fhSingle, (void *)fm82c0 },
};

//===	83

static const FONTMAP	fm8340[] = {
	{ 0x40, 0x7e, 0x0124 },
};

static const FONTMAP	fm8380[] = {
	{ 0x80, 0x96, 0x0163 },
	{ 0x9f, 0xb6, 0x017a },
	{ 0xbf, 0xbf, 0x0192 },
};
static const MFONTMAP	mfm8380[] = { ARRAYSIZE(fm8380), (PFONTMAP)fm8380 };

static const FONTMAP	fm83c0[] = {
	{ 0xc0, 0xd6, 0x0193 },
};

static const FONTHAND	fh83[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fm8340 },
	{ fhMulti, (void *)mfm8380 },
	{ fhSingle, (void *)fm83c0 },
};

//===	84

static const FONTMAP	fm8440[] = {
	{ 0x40, 0x60, 0x01aa },
	{ 0x70, 0x7e, 0x01cb },
};
static const MFONTMAP	mfm8440[] = { ARRAYSIZE(fm8440), (PFONTMAP)fm8440 };

static const FONTMAP	fm8480[] = {
	{ 0x80, 0x91, 0x01da },
	{ 0x9f, 0xbe, 0x01ec },
};
static const MFONTMAP	mfm8480[] = { ARRAYSIZE(fm8480), (PFONTMAP)fm8480 };

static const FONTHAND	fh84[] = {
	{ fhInvalid, NULL },
	{ fhMulti, (void *)mfm8440 },
	{ fhMulti, (void *)mfm8480 },
	{ fhInvalid, NULL },
};

//===	88

static const FONTMAP	fm8880[] = {
	{ 0x9f, 0xbf, 0x020c },
};

static const FONTMAP	fm88c0[] = {
	{ 0xc0, 0xfc, 0x022d },
};

static const FONTHAND	fh88[] = {
	{ fhInvalid, NULL },
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fm8880 },
	{ fhSingle, (void *)fm88c0 },
};

//===	89

static const FONTMAP	fm8940[] = {
	{ 0x40, 0x7e, 0x026a },
};

static const FONTMAP	fm8980[] = {
	{ 0x80, 0xfc, 0x02a9 },
};

static const FONTHAND	fh89[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fm8940 },
	{ fhSingle, (void *)fm8980 },
	{ fhSingle, (void *)fm8980 },
};

//===	8a

static const FONTMAP	fm8a40[] = {
	{ 0x40, 0x7e, 0x0326 },
};

static const FONTMAP	fm8a80[] = {
	{ 0x80, 0xfc, 0x0365 },
};

static const FONTHAND	fh8a[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fm8a40 },
	{ fhSingle, (void *)fm8a80 },
	{ fhSingle, (void *)fm8a80 },
};

//===	8b

static const FONTMAP	fm8b40[] = {
	{ 0x40, 0x7e, 0x03e2 },
};

static const FONTMAP	fm8b80[] = {
	{ 0x80, 0xfc, 0x0421 },
};

static const FONTHAND	fh8b[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fm8b40 },
	{ fhSingle, (void *)fm8b80 },
	{ fhSingle, (void *)fm8b80 },
};

//===	8c

static const FONTMAP	fm8c40[] = {
	{ 0x40, 0x7e, 0x049e },
};

static const FONTMAP	fm8c80[] = {
	{ 0x80, 0xfc, 0x04dd },
};

static const FONTHAND	fh8c[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fm8c40 },
	{ fhSingle, (void *)fm8c80 },
	{ fhSingle, (void *)fm8c80 },
};

//===	8d

static const FONTMAP	fm8d40[] = {
	{ 0x40, 0x7e, 0x055a },
};

static const FONTMAP	fm8d80[] = {
	{ 0x80, 0xfc, 0x0599 },
};

static const FONTHAND	fh8d[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fm8d40 },
	{ fhSingle, (void *)fm8d80 },
	{ fhSingle, (void *)fm8d80 },
};

//===	8e

static const FONTMAP	fm8e40[] = {
	{ 0x40, 0x7e, 0x0616 },
};

static const FONTMAP	fm8e80[] = {
	{ 0x80, 0xfc, 0x0655 },
};

static const FONTHAND	fh8e[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fm8e40 },
	{ fhSingle, (void *)fm8e80 },
	{ fhSingle, (void *)fm8e80 },
};

//===	8f

static const FONTMAP	fm8f40[] = {
	{ 0x40, 0x7e, 0x06d2 },
};

static const FONTMAP	fm8f80[] = {
	{ 0x80, 0xfc, 0x0711 },
};

static const FONTHAND	fh8f[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fm8f40 },
	{ fhSingle, (void *)fm8f80 },
	{ fhSingle, (void *)fm8f80 },
};

//===	90

static const FONTMAP	fm9040[] = {
	{ 0x40, 0x7e, 0x078e },
};

static const FONTMAP	fm9080[] = {
	{ 0x80, 0xfc, 0x07cd },
};

static const FONTHAND	fh90[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fm9040 },
	{ fhSingle, (void *)fm9080 },
	{ fhSingle, (void *)fm9080 },
};

//===	91

static const FONTMAP	fm9140[] = {
	{ 0x40, 0x7e, 0x084a },
};

static const FONTMAP	fm9180[] = {
	{ 0x80, 0xfc, 0x0889 },
};

static const FONTHAND	fh91[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fm9140 },
	{ fhSingle, (void *)fm9180 },
	{ fhSingle, (void *)fm9180 },
};

//===	92

static const FONTMAP	fm9240[] = {
	{ 0x40, 0x7e, 0x0906 },
};

static const FONTMAP	fm9280[] = {
	{ 0x80, 0xfc, 0x0945 },
};

static const FONTHAND	fh92[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fm9240 },
	{ fhSingle, (void *)fm9280 },
	{ fhSingle, (void *)fm9280 },
};

//===	93

static const FONTMAP	fm9340[] = {
	{ 0x40, 0x7e, 0x09c2 },
};

static const FONTMAP	fm9380[] = {
	{ 0x80, 0xfc, 0x0a01 },
};

static const FONTHAND	fh93[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fm9340 },
	{ fhSingle, (void *)fm9380 },
	{ fhSingle, (void *)fm9380 },
};

//===	94

static const FONTMAP	fm9440[] = {
	{ 0x40, 0x7e, 0x0a7e },
};

static const FONTMAP	fm9480[] = {
	{ 0x80, 0xfc, 0x0abd },
};

static const FONTHAND	fh94[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fm9440 },
	{ fhSingle, (void *)fm9480 },
	{ fhSingle, (void *)fm9480 },
};

//===	95

static const FONTMAP	fm9540[] = {
	{ 0x40, 0x7e, 0x0b3a },
};

static const FONTMAP	fm9580[] = {
	{ 0x80, 0xfc, 0x0b79 },
};

static const FONTHAND	fh95[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fm9540 },
	{ fhSingle, (void *)fm9580 },
	{ fhSingle, (void *)fm9580 },
};

//===	96

static const FONTMAP	fm9640[] = {
	{ 0x40, 0x7e, 0x0bf6 },
};

static const FONTMAP	fm9680[] = {
	{ 0x80, 0xfc, 0x0c35 },
};

static const FONTHAND	fh96[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fm9640 },
	{ fhSingle, (void *)fm9680 },
	{ fhSingle, (void *)fm9680 },
};

//===	97

static const FONTMAP	fm9740[] = {
	{ 0x40, 0x7e, 0x0cb2 },
};

static const FONTMAP	fm9780[] = {
	{ 0x80, 0xfc, 0x0cf1 },
};

static const FONTHAND	fh97[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fm9740 },
	{ fhSingle, (void *)fm9780 },
	{ fhSingle, (void *)fm9780 },
};

//===	98

static const FONTMAP	fm9840[] = {
	{ 0x40, 0x72, 0x0d6e },
};

static const FONTMAP	fm9880[] = {
	{ 0x9f, 0xfc, 0x0da1 },
};

static const FONTHAND	fh98[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fm9840 },
	{ fhSingle, (void *)fm9880 },
	{ fhSingle, (void *)fm9880 },
};

//===	99

static const FONTMAP	fm9940[] = {
	{ 0x40, 0x7e, 0x0dff },
};

static const FONTMAP	fm9980[] = {
	{ 0x80, 0xfc, 0x0e3e },
};

static const FONTHAND	fh99[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fm9940 },
	{ fhSingle, (void *)fm9980 },
	{ fhSingle, (void *)fm9980 },
};

//===	9a

static const FONTMAP	fm9a40[] = {
	{ 0x40, 0x7e, 0x0ebb },
};

static const FONTMAP	fm9a80[] = {
	{ 0x80, 0xfc, 0x0efa },
};

static const FONTHAND	fh9a[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fm9a40 },
	{ fhSingle, (void *)fm9a80 },
	{ fhSingle, (void *)fm9a80 },
};

//===	9b

static const FONTMAP	fm9b40[] = {
	{ 0x40, 0x7e, 0x0f77 },
};

static const FONTMAP	fm9b80[] = {
	{ 0x80, 0xfc, 0x0fb6 },
};

static const FONTHAND	fh9b[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fm9b40 },
	{ fhSingle, (void *)fm9b80 },
	{ fhSingle, (void *)fm9b80 },
};

//===	9c

static const FONTMAP	fm9c40[] = {
	{ 0x40, 0x7e, 0x1033 },
};

static const FONTMAP	fm9c80[] = {
	{ 0x80, 0xfc, 0x1072 },
};

static const FONTHAND	fh9c[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fm9c40 },
	{ fhSingle, (void *)fm9c80 },
	{ fhSingle, (void *)fm9c80 },
};

//===	9d

static const FONTMAP	fm9d40[] = {
	{ 0x40, 0x7e, 0x10ef },
};

static const FONTMAP	fm9d80[] = {
	{ 0x80, 0xfc, 0x112e },
};

static const FONTHAND	fh9d[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fm9d40 },
	{ fhSingle, (void *)fm9d80 },
	{ fhSingle, (void *)fm9d80 },
};

//===	9e

static const FONTMAP	fm9e40[] = {
	{ 0x40, 0x7e, 0x11ab },
};

static const FONTMAP	fm9e80[] = {
	{ 0x80, 0xfc, 0x11ea },
};

static const FONTHAND	fh9e[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fm9e40 },
	{ fhSingle, (void *)fm9e80 },
	{ fhSingle, (void *)fm9e80 },
};

//===	9f

static const FONTMAP	fm9f40[] = {
	{ 0x40, 0x7e, 0x1267 },
};

static const FONTMAP	fm9f80[] = {
	{ 0x80, 0xfc, 0x12a6 },
};

static const FONTHAND	fh9f[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fm9f40 },
	{ fhSingle, (void *)fm9f80 },
	{ fhSingle, (void *)fm9f80 },
};

//===	low

static const PFONTHAND	fhlo[] = {
	(PFONTHAND)fh81,
	(PFONTHAND)fh82,
	(PFONTHAND)fh83,
	(PFONTHAND)fh84,
	NULL,			// 85
	NULL,			// 86
	NULL,			// 87
	(PFONTHAND)fh88,
	(PFONTHAND)fh89,
	(PFONTHAND)fh8a,
	(PFONTHAND)fh8b,
	(PFONTHAND)fh8c,
	(PFONTHAND)fh8d,
	(PFONTHAND)fh8e,
	(PFONTHAND)fh8f,
	(PFONTHAND)fh90,
	(PFONTHAND)fh91,
	(PFONTHAND)fh92,
	(PFONTHAND)fh93,
	(PFONTHAND)fh94,
	(PFONTHAND)fh95,
	(PFONTHAND)fh96,
	(PFONTHAND)fh97,
	(PFONTHAND)fh98,
	(PFONTHAND)fh99,
	(PFONTHAND)fh9a,
	(PFONTHAND)fh9b,
	(PFONTHAND)fh9c,
	(PFONTHAND)fh9d,
	(PFONTHAND)fh9e,
	(PFONTHAND)fh9f,
};

//===	e0

static const FONTMAP	fme040[] = {
	{ 0x40, 0x7e, 0x1323 },
};

static const FONTMAP	fme080[] = {
	{ 0x80, 0xfc, 0x1362 },
};

static const FONTHAND	fhe0[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fme040 },
	{ fhSingle, (void *)fme080 },
	{ fhSingle, (void *)fme080 },
};

//===	e1

static const FONTMAP	fme140[] = {
	{ 0x40, 0x7e, 0x13df },
};

static const FONTMAP	fme180[] = {
	{ 0x80, 0xfc, 0x141e },
};

static const FONTHAND	fhe1[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fme140 },
	{ fhSingle, (void *)fme180 },
	{ fhSingle, (void *)fme180 },
};

//===	e2

static const FONTMAP	fme240[] = {
	{ 0x40, 0x7e, 0x149b },
};

static const FONTMAP	fme280[] = {
	{ 0x80, 0xfc, 0x14da },
};

static const FONTHAND	fhe2[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fme240 },
	{ fhSingle, (void *)fme280 },
	{ fhSingle, (void *)fme280 },
};

//===	e3

static const FONTMAP	fme340[] = {
	{ 0x40, 0x7e, 0x1557 },
};

static const FONTMAP	fme380[] = {
	{ 0x80, 0xfc, 0x1596 },
};

static const FONTHAND	fhe3[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fme340 },
	{ fhSingle, (void *)fme380 },
	{ fhSingle, (void *)fme380 },
};

//===	e4

static const FONTMAP	fme440[] = {
	{ 0x40, 0x7e, 0x1613 },
};

static const FONTMAP	fme480[] = {
	{ 0x80, 0xfc, 0x1652 },
};

static const FONTHAND	fhe4[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fme440 },
	{ fhSingle, (void *)fme480 },
	{ fhSingle, (void *)fme480 },
};

//===	e5

static const FONTMAP	fme540[] = {
	{ 0x40, 0x7e, 0x16cf },
};

static const FONTMAP	fme580[] = {
	{ 0x80, 0xfc, 0x170e },
};

static const FONTHAND	fhe5[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fme540 },
	{ fhSingle, (void *)fme580 },
	{ fhSingle, (void *)fme580 },
};

//===	e6

static const FONTMAP	fme640[] = {
	{ 0x40, 0x7e, 0x178b },
};

static const FONTMAP	fme680[] = {
	{ 0x80, 0xfc, 0x17ca },
};

static const FONTHAND	fhe6[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fme640 },
	{ fhSingle, (void *)fme680 },
	{ fhSingle, (void *)fme680 },
};

//===	e7

static const FONTMAP	fme740[] = {
	{ 0x40, 0x7e, 0x1847 },
};

static const FONTMAP	fme780[] = {
	{ 0x80, 0xfc, 0x1886 },
};

static const FONTHAND	fhe7[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fme740 },
	{ fhSingle, (void *)fme780 },
	{ fhSingle, (void *)fme780 },
};

//===	e8

static const FONTMAP	fme840[] = {
	{ 0x40, 0x7e, 0x1903 },
};

static const FONTMAP	fme880[] = {
	{ 0x80, 0xfc, 0x1942 },
};

static const FONTHAND	fhe8[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fme840 },
	{ fhSingle, (void *)fme880 },
	{ fhSingle, (void *)fme880 },
};

//===	e9

static const FONTMAP	fme940[] = {
	{ 0x40, 0x7e, 0x19bf },
};

static const FONTMAP	fme980[] = {
	{ 0x80, 0xfc, 0x19fe },
};

static const FONTHAND	fhe9[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fme940 },
	{ fhSingle, (void *)fme980 },
	{ fhSingle, (void *)fme980 },
};

//===	ea

static const FONTMAP	fmea40[] = {
	{ 0x40, 0x7e, 0x1a7b },
};

static const FONTMAP	fmea80[] = {
	{ 0x80, 0xa2, 0x1aba },
};

static const FONTHAND	fhea[] = {
	{ fhInvalid, NULL },
	{ fhSingle, (void *)fmea40 },
	{ fhSingle, (void *)fmea80 },
	{ fhSingle, (void *)fmea80 },
};

//===	high

static const PFONTHAND	fhhi[] = {
	(PFONTHAND)fhe0,
	(PFONTHAND)fhe1,
	(PFONTHAND)fhe2,
	(PFONTHAND)fhe3,
	(PFONTHAND)fhe4,
	(PFONTHAND)fhe5,
	(PFONTHAND)fhe6,
	(PFONTHAND)fhe7,
	(PFONTHAND)fhe8,
	(PFONTHAND)fhe9,
	(PFONTHAND)fhea,
};

/*
 *	externs
 */

extern const u8	ElisaFont[];

/*
 *	Code2FontOffset
 */
static int Code2FontOffset(u16 code)
{
	u8		ch, cl;
	u8		ix;
	PFONTHAND	pfh;
	int		index;

	ch = HIBYTE(code);
	if (ch >= 0x81 && ch <= 0x9f)
		pfh = fhlo[ch - 0x81];
	else if (ch >= 0xe0 && ch <= 0xea)
		pfh = fhhi[ch - 0xe0];
	else
		return INVALID_OFFSET;

	if (!pfh)
		return INVALID_OFFSET;

	cl = LOBYTE(code);
	ix = cl >> 6;
	pfh = &pfh[ix];
	index = (pfh->handler)(cl, pfh->pdata);

	return INDEX2OFFSET(index);
}

static u16	fgColor, bgColor;

/*
 *	ElisaSetColor
 */
void ElisaSetColor(int fgcolor, int bgcolor)
{
	fgColor = fgcolor;
	bgColor = bgcolor;
}

/*
 *	ElisaGetCharWidth
 */
int ElisaGetCharWidth(u16 code)
{
	return FONT_WIDTH;
}

/*
 *	ElisaPutChar
 */
int ElisaPutChar(int x, int y, u16 code)
{
	vu16		*v;
	u8		*fp;
	int		i;
	u8		fm;
	u8		fd;

	fp = (u8 *)&ElisaFont[Code2FontOffset(code)];
	v = (vu16 *)VRAM_ADDR + y * SCREEN_WIDTH + x;
	for (i = 0; i < 8; i++) {
		fd = *fp++;
		for (fm = 0x80; fm; fm >>= 1)
			*v++ = (fd & fm) ? fgColor : bgColor;
		*v = bgColor;
		v += (SCREEN_WIDTH - 8);
	}

	return FONT_WIDTH;
}

/*
 *	ElisaPutChar2
 */
int ElisaPutChar2(int x, int y, u16 code)
{
	vu16		*v;
	u8		*fp;
	int		i;
	u8		fm;
	u8		fd;
	u16		bm;

	fp = (u8 *)&ElisaFont[Code2FontOffset(code)];
	v = (vu16 *)VRAM_ADDR + y * SCREEN_WIDTH + x;
	for (i = 0; i < 8; i++) {
		fd = *fp++;
		for (fm = 0x80; fm; fm >>= 1) {
			bm = (fd & fm) ? fgColor : bgColor;
			v[SCREEN_WIDTH] = bm;
			*v++ = bm;
			v[SCREEN_WIDTH] = bm;
			*v++ = bm;
		}
		v[SCREEN_WIDTH] = bgColor;
		*v++ = bgColor;
		v[SCREEN_WIDTH] = bgColor;
		*v++ = bgColor;
		v += (SCREEN_WIDTH - 9*2 + SCREEN_WIDTH);
	}

	return FONT_WIDTH * 2;
}

