/*
 *	vsprintf.c
 */

#include <gba.h>
#include <stdarg.h>

#include <font8.h>

typedef struct _VALFORM {
	s16	base;
	u8	sign;
	u8	len;
	u8	fc;
	u8	comma;
} VALFORM, *PVALFORM;

static const u8	ItoA[] = "0123456789ABCDEF";

/*
 *	PutValue
 */
static u8 *PutValue(u8 *bp, s32 v, PVALFORM pvf)
{
	u32		uv;
	u8		*vp;
	u8		n, c;
	u8		vbuf[32];

	if (pvf->sign && v < 0) {
		*bp++ = '-';
		uv = (u32)-v;
	} else
		uv = v;

	vp = vbuf;
	n = 0;
	c = 0;
	do {
		if (pvf->comma && c && (c % 3) == 0) {
			*vp++ = ',';
			n++;
		}
		*vp++ = ItoA[uv % pvf->base];
		uv /= pvf->base;
		n++;
		c++;
	} while (uv > 0);
	if (pvf->len)
		while (n < pvf->len && vp < &vbuf[16]) {
			*vp++ = pvf->fc;
			n++;
		}
	while (vp > vbuf)
		*bp++ = *--vp;

	return bp;
}

/*
 *	_vsprintf
 */
void _vsprintf(u8 *bp, u8 *fmt, va_list vp)
{
	u8		c;
	s32		v;
	u8		*s;
	VALFORM		vf;

	while (c = *fmt++) {
		if (c != '%') {
			*bp++ = c;
			continue;
		}

		vf.sign = TRUE;
		vf.len = 0;
		vf.fc = ' ';
		vf.comma = FALSE;

	format:
		switch (c = *fmt++) {
		default:
			*bp++ = c;
			break;
		case '0':
			if (!vf.len) {
				vf.fc = c;
				goto format;
			}
			// FALL THRU
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			vf.len = vf.len * 10 + (c - '0');
			goto format;
		case ',':
			vf.comma = TRUE;
			goto format;
		case 'd':
			v = va_arg(vp, s32);
			vf.base = 10;
			bp = PutValue(bp, v, &vf);
			break;
		case 'x':
			v = va_arg(vp, u32);
			vf.base = 16;
			bp = PutValue(bp, v, &vf);
			break;
		case 's':
			s = va_arg(vp, u8 *);
			while (*s)
				*bp++ = *s++;
			break;
		}
	}

	*bp++ = '\0';
}

/*
 *	_sprintf
 */
void _sprintf(u8 *bp, u8 *fmt, ...)
{
	va_list		vp;

	va_start(vp, fmt);
	_vsprintf(bp, fmt, vp);
	va_end(vp);
}

