/*
 *	gbafs.c
 */

#include <gbafs.h>
#include <chartype.h>

#define ROUNDUPFS(x)	(((x) + (ROMFS_CHUNK - 1)) & ~(ROMFS_CHUNK - 1))

static PROMFSHEAD	pRomFsHead;
static PROMDIRENT	pRootDir;

extern int	__rodata_end;

/*
 *	IsValidFs
 */
static BOOL IsValidFs()
{
	if (!pRomFsHead ||
		memcmp(pRomFsHead->signature, ROMFSSIG, sizeof(ROMFSSIG)-1) ||
		pRomFsHead->version != 0x01)
		return FALSE;

	return TRUE;
}

/*
 *	FsInit
 */
BOOL FsInit()
{
	pRomFsHead = (PROMFSHEAD)ROUNDUPFS((u32)&__rodata_end);
	if (!IsValidFs())
		return FALSE;

	pRootDir = &pRomFsHead->rootDir;

	return TRUE;
}

/*
 *	FSGetRootDir
 */
PDIRENT FsGetRootDir()
{
	if (!IsValidFs())
		return FALSE;

	return (PDIRENT)pRootDir;
}

/*
 *	FsGetDirEnt
 */
PDIRENT FsGetDirEnt(u8 *path)
{
	PDIRENT		pd, pe, pdir;
	u8		*p0, *p;
	u8		c;
	int		len;

	pdir = NULL;
	p0 = NULL;
	for (;;) {
		switch (c = *path) {
		default:
			if (!p0)
				p0 = path;
			CHARNEXT(&path);
			break;
		case '\\':
		case '\0':
			if (!p0) {
				if (!pdir) {
					pdir = FsGetRootDir();
					if (!pdir)
						goto out;
				}
				if (!c)
					goto out;
				path++;
				goto next;
			}

			if (!pdir)
				goto out;

			len = path - p0;
			pd = (PDIRENT)(pdir->fileaddr + pdir->namesize);
			pe = (PDIRENT)((u8 *)pd + pdir->filesize);
			for (; pd < pe; pd++) {
				p = pd->fileaddr;
				if (strncmp(p, p0, len) == 0 && p[len] == 0) {
					pdir = pd;
					if (!c)
						goto out;
					path++;
					p0 = NULL;
					goto next;
				}
			}
			pdir = NULL;
			goto out;
		}
	next:
		;
	}

out:
	return pdir;
}

