/*
 *	xfont.c
 */

#include <xfont.h>

static const u8	BitMasks[] = { 0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01 };

static u16	fgColor, bgColor;

/*
 *	IsValidXfont
 */
BOOL IsValidXfont(PXFONTHEAD pxf)
{
	return memcmp(pxf->sig, XFONT_SIG, sizeof(pxf->sig)) == 0 &&
		pxf->ver == XFONT_VER &&
		pxf->size == sizeof(*pxf);
}

/*
 *	XfontSetColor
 */
void XfontSetColor(int fgcolor, int bgcolor)
{
	fgColor = fgcolor;
	bgColor = bgcolor;
}

/*
 *	XfontGetFontName
 */
u8 *XfontGetFontName()
{
	return XfontGetFontObject(' ')->pxf->name;
}

/*
 *	XfontGetMaxCharWidth
 */
int XfontGetMaxCharWidth(u8 code)
{
	PXFONTHEAD	pxf;

	pxf = XfontGetFontObject(code)->pxf;
	return pxf->width + pxf->xgap;
}

/*
 *	XfontGetCharHeight
 */
int XfontGetCharHeight()
{
	PXFONTHEAD	pxf;

	pxf = XfontGetFontObject(' ')->pxf;
	return pxf->height + pxf->ygap;
}

/*
 *	XfontGetCharWidth
 */
int XfontGetCharWidth(u8 **ppcode)
{
	PXFONTOBJ	pxo;
	PXFONTHEAD	pxf;
	int		off;
	u8		*pf;
	u16		code;

	pxo = XfontGetFontObject(**ppcode);
	pxf = pxo->pxf;
	code = pxo->getcode(ppcode);
	if (pxf->flags & XFF_PROP) {
		off = pxo->code2index(pxf, code) * pxf->unit;
		pf = ((u8 *)pxf) + pxf->size + off;

		return *pf + pxf->xgap;
	} else
		return pxf->width + pxf->xgap;
}

/*
 *	XfontPutChar
 */
int XfontPutChar(int x, int y, u8 **ppcode)
{
	PXFONTOBJ	pxo;
	PXFONTHEAD	pxf;
	vu16		*v;
	u8		*pf;
	int		off;
	int		i, j;
	int		width;
	u8		bmask;
	u8		fd;
	u16		code;

	pxo = XfontGetFontObject(**ppcode);
	pxf = pxo->pxf;
	code = pxo->getcode(ppcode);
	if (pxf->flags & XFF_PROP) {
		off = pxo->code2index(pxf, code) * pxf->unit;
		pf = ((u8 *)pxf) + pxf->size + off;
		width = *pf++;
		bmask = 0x80;
	} else {
		off = pxo->code2index(pxf, code) * pxf->unit;
		pf = ((u8 *)pxf) + pxf->size + (off / NBPB);
		width = pxf->width;
		bmask = BitMasks[off % NBPB];
	}

	v = (vu16 *)VRAM_ADDR + y * SCREEN_WIDTH + x;
	fd = *pf++;
	for (i = 0; i < pxf->height; i++) {
		for (j = 0; j < width; j++) {
			*v++ = (fd & bmask) ? fgColor : bgColor;
			if ((bmask >>= 1) == 0) {
				bmask = 0x80;
				fd = *pf++;
			}
		}
		for (j = 0; j < pxf->xgap; j++)
			*v++ = bgColor;
		v += SCREEN_WIDTH - (width + pxf->xgap);
	}

	for (i = 0; i < pxf->ygap; i++) {
		for (j = 0; j < width + pxf->xgap; j++)
			*v++ = bgColor;
		v += SCREEN_WIDTH - (width + pxf->xgap);
	}

	return width + pxf->xgap;
}

